#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include <time.h>

#define N_ltc 30  // it should be equal to sqrt(N_eig)
#define N_par 900
#define N_eig 900

double res[N_eig][N_par]; 
double ham[N_eig][N_eig];

void generate_matrix(double parameter)
{
   double alpha,x,y;
   int i,j;

   alpha=parameter*2*3.141592653589793;
   for (i=0;i<N_eig;i++)
     for (j=0;j<N_eig;j++) ham[i][j]=0;
   for (i=1;i<N_eig;i++) {
     x=(i-1)%N_ltc;
     y=(i-1)/N_ltc;
     if(x<N_ltc-1) {
       ham[i-1][i]=cos(-alpha*y/2);
       ham[i][i-1]=cos(alpha*y/2);
     }
     if (y<N_ltc-1) {
       ham[i-1][i+N_ltc-1]=cos(alpha*x/2);
       ham[i+N_ltc-1][i-1]=cos(-alpha*x/2);  
     }
   }

}


int main (int argc, char **argv)
{
    int rank, nprocs, tag,N_cal;
    int i, j, k,n, start, end,desca[9];
    unsigned long *val;
    unsigned long sum, total;
    int iter,ione,izero,info,N=N_eig;
    int cnt=0;
    FILE *fp, *fp2;
    MPI_Status status;
    MPI_Request req1, req2;
    double interval,startparameter;
    double localtime,etime,work[2570000];
    int mype,npe,context,map[2],lwork;
    char jobz,ulo;
    if(argc!=2) {
      printf("filename\n");
      return 0;
    } 


    MPI_Init (&argc, &argv);
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);
    MPI_Comm_size (MPI_COMM_WORLD, &nprocs);

    Cblacs_pinfo( &mype, &npe);
    Cblacs_get(-1,0,&context);
    Cblacs_gridinit(&context,"R",1,npe);
    //printf("Context = %d , mype = %d, npe = %d\n",context,mype,npe); 
   
    map[0]=rank;
    Cblacs_gridmap(&context,map,1,1,1);
    ione=1;izero=0; lwork=2570000;
    jobz='N';ulo='U'; N=N_eig;
    
    descinit_(desca,&N,&N,&N,&N,&izero,&izero,&context,&N,&info); 
    

    interval=1./N_par;
    //startparameter = (double) (rank-1)*((double) 1/(nprocs-1));
    //iter = (int)((1./(nprocs-1.)) / interval);
    startparameter = ((double) (1.0/nprocs)) * rank;
    iter=N_par/nprocs; 
    printf("iter = %d\n",iter);
    double temp[iter*N_eig],eigen[iter*N_eig]; 

    N_cal=iter*(nprocs-1);
    
    if (rank==0) {
      etime=(double) clock();  
      
      printf("Start!\n");
      for (i=0;i<iter;i++) {
        generate_matrix(i*interval);
        pdsyev_(&jobz,&ulo,&N,ham,&ione,&ione,desca,eigen+i*N,NULL,&ione,&ione,NULL,work,&lwork,&info);
      //for (j=0;j<N_eig;j++) res[j][i]=eigen[j];
     }
     for (i=0;i<iter;i++)
       for (j=0;j<N_eig;j++)
         res[j][i]=eigen[j+i*N_eig]; 

   
     printf("end! Overheadtime = %f\n",((double) clock() - etime)/(double) CLOCKS_PER_SEC);
     cnt=0;
     while(cnt<nprocs-1) {
       MPI_Recv(temp,N*iter,MPI_DOUBLE,cnt+1,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
       for (i=0;i<iter;i++)
          for (j=0;j<N_eig;j++)
            res[j][i+(cnt+1)*iter]=temp[j+i*N_eig];
       cnt++;
     }
     fp = fopen(argv[1],"w");
     for (i=0;i<N_eig;i++){
       for (j=0;j<N_par;j++)
         fprintf(fp,"%f ",res[i][j]);
       fprintf(fp,"\n");
     }
     fclose(fp);
     etime = ((double) clock() - etime )/(double) CLOCKS_PER_SEC; 
     fp2 = fopen("mpi_time.dat","w");
     fprintf(fp2,"mpi_time = %1f\n",etime);
     fclose(fp2); 
   }
   else {
      

      localtime = (double) clock();
      for (i=0;i<iter;i++) {
        generate_matrix(startparameter+interval*i);
        pdsyev_(&jobz,&ulo,&N,ham,&ione,&ione,desca,eigen+i*N,NULL,&ione,&ione,NULL,work,&lwork,&info);
       }
       localtime = ((double) clock() - localtime )/(double) CLOCKS_PER_SEC;
       printf("elapsed time for rank %d is %f\n",rank,localtime);
       MPI_Send(eigen,N*iter,MPI_DOUBLE,0,rank,MPI_COMM_WORLD); 
    }
 
    Cblacs_gridexit(context);

    MPI_Finalize ();
}

